#!/usr/bin/env python3

"""
(C) 2022,2023,2025 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import hashlib
import json
import multiprocessing
from multiprocessing.pool import ThreadPool
import optparse # pylint: disable=deprecated-module
import os
import random
import re
import shlex
import sqlite3
import subprocess
import sys
import time
import uuid

enabled_checks = [
    'bugprone-*',
    'cert-*',
    'clang-analyzer-*',
    'cppcoreguidelines-*',
    'hicpp-*',
    'misc-*',
    'modernize-*',
    'performance-*',
    'portability-*',
    'readability-*',
]

# these are ones that we might want to be clean for in the future,
# but currently are not
disabled_needs_work = [
    'cppcoreguidelines-use-default-member-init',
    'bugprone-unchecked-optional-access',
    'bugprone-empty-catch',
    'cppcoreguidelines-avoid-const-or-ref-data-members',
    'cppcoreguidelines-pro-type-const-cast',
    'misc-include-cleaner', # warning: useful but quite buggy
    'misc-multiple-inheritance', # public key uses this but should be fixed
    'misc-override-with-different-visibility', # mostly fine but should be looked at
    'modernize-pass-by-value',
    'modernize-use-ranges', # limited by compiler support currently
    'performance-avoid-endl',
    'readability-redundant-typename', # TODO(Botan4) when Clang min version increases this can be fixed
    'readability-convert-member-functions-to-static',
    'readability-inconsistent-declaration-parameter-name', # should fix this, blocked by https://github.com/llvm/llvm-project/issues/60845
    'readability-simplify-boolean-expr', # sometimes ok
    'readability-static-accessed-through-instance',
]

# these we are probably not interested in ever being clang-tidy clean for
disabled_not_interested = [
    '*-array-to-pointer-decay',
    '*-avoid-c-arrays', # triggers also on foo(T x[], size_t len) decls
    '*-else-after-return',
    '*-function-size',
    '*-magic-numbers', # can't stop the magic
    '*-narrowing-conversions',
    '*-no-array-decay',
    '*-use-auto', # not universally a good idea
    '*-use-emplace', # often less clear
    '*-deprecated-headers', # wrong for system headers like stdlib.h
    'cert-dcl21-cpp', # invalid, and removed already in clang-tidy 19
    'bugprone-easily-swappable-parameters',
    'bugprone-implicit-widening-of-multiplication-result', # would be reasonable if it ignored constants (it doesn't)
    'cppcoreguidelines-pro-bounds-avoid-unchecked-container-access',
    'cppcoreguidelines-pro-bounds-pointer-arithmetic',
    'cppcoreguidelines-pro-bounds-constant-array-index',
    'cppcoreguidelines-pro-type-reinterpret-cast',
    'hicpp-signed-bitwise', # would be reasonable if it ignored constants (it doesn't)
    'misc-no-recursion',
    'modernize-avoid-c-style-cast', # the kind of cast clang-tidy is complaining about here is fine
    'modernize-use-trailing-return-type', # fine, but we're not using it everywhere
    'modernize-return-braced-init-list', # thanks I hate it
    'modernize-use-default-member-init',
    'modernize-use-designated-initializers',
    'modernize-use-nodiscard',
    'modernize-use-using', # fine not great
    'readability-avoid-return-with-void-value', # Jack likes doing this
    'readability-function-cognitive-complexity',
    'readability-identifier-length', # lol, lmao
    'readability-math-missing-parentheses',
    'readability-non-const-parameter',
    'readability-use-concise-preprocessor-directives', # it's not more readable...
    'readability-redundant-parentheses', # often improves readability ...
    'readability-redundant-inline-specifier', # Jack likes doing this
    'readability-redundant-access-specifiers', # reneme likes doing this
    'readability-use-anyofallof', # not more readable
]

disabled_checks = sorted(disabled_needs_work + disabled_not_interested)

def create_check_option(enabled, disabled):
    return ','.join(enabled) + ',' + ','.join(['-' + d for d in disabled])

def render_clang_tidy_file(target_dir, enabled, disabled):
    filepath = os.path.join(target_dir, '.clang-tidy')
    print(f'regenerating {filepath}')
    with open(filepath, "w", encoding="utf-8") as clang_tidy_file:
        clang_tidy_file.writelines([
            '---\n',
            f'# This file was automatically generated by {sys.argv[0]} --regenerate-inline-config-file\n',
            '#\n',
            '# All manual edits to this file will be lost. Edit the script\n',
            '# then regenerate this configuration file.\n',
            '\n',
            'Checks: >\n'] +
            [ f'    {check},\n' for check in sorted(enabled)] +
            [ f'    -{check},\n' for check in sorted(disabled)] +
            ['---\n'])

def load_compile_commands(build_dir):
    compile_commands_file = os.path.join(build_dir, 'compile_commands.json')
    compile_commands = open(compile_commands_file, encoding='utf8').read()
    return (compile_commands_file, json.loads(compile_commands))

def run_command(cmdline):
    proc = subprocess.Popen(cmdline,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    (stdout, _stderr) = proc.communicate()

    stdout = stdout.decode('utf8')
    # stderr is discarded

    return stdout

def run_clang_tidy(compile_commands_file,
                   check_config,
                   source_file,
                   options):

    cmdline = ['clang-tidy',
               '--quiet',
               '-checks=%s' % (check_config),
               '-p', compile_commands_file,
               '-header-filter=.*']

    if options.fixit:
        cmdline.append('-fix')

    if options.export_fixes_dir:
        os.makedirs(options.export_fixes_dir, exist_ok=True)
        yml_file = os.path.join(options.export_fixes_dir, "%s.yml" % uuid.uuid4())
        cmdline.append("--export-fixes=%s" % yml_file)

    cmdline.append(source_file)

    stdout = run_command(cmdline)

    if stdout != "":
        print("### Errors in", source_file)
        print(stdout)
        sys.stdout.flush()
        return False

    return True

def file_matches(file, args):
    if args is None or len(args) == 0:
        return True

    for arg in args:
        if file.find(arg) > 0:
            return True
    return False

def preproc_file(compile_commands):
    cmd = shlex.split(compile_commands['command'])

    dash_o = cmd.index("-o")
    if dash_o >= 0:
        cmd.pop(dash_o + 1) # remove object file name
        cmd.pop(dash_o)     # remove -o

    if "-c" in cmd:
        cmd.remove("-c")
    cmd.append("-E")

    result = subprocess.run(
        cmd,
        check=True,
        stdout=subprocess.PIPE,
        universal_newlines=True)

    return hashlib.sha256(result.stdout.encode('utf-8')).hexdigest()

def hash_args(**kwargs):
    outer_hash = hashlib.sha256()

    for key in sorted(kwargs.keys()):
        value = kwargs[key]
        outer_hash.update(hashlib.sha256(key.encode('utf-8')).digest())
        outer_hash.update(hashlib.sha256(value.encode('utf-8')).digest())

    return outer_hash.hexdigest()

class CacheDatabase:
    """
    Caches SUCCESSFUL clang-tidy runs on source files to speed up whole-tree
    runs of clang-tidy. The cache key is calculated as the hash of the source file
    and a set of meta information such as 'clang-tidy --version' and the check configuration.
    """
    CACHE_DEBUG = False

    def __init__(self, db_path):
        self.db = sqlite3.connect(db_path)
        cur = self.db.cursor()
        cur.execute("CREATE TABLE IF NOT EXISTS clang_tidy_clean(key UNIQUE)")
        self.db.commit()

    def hit(self, **kwargs):
        cache_key = hash_args(**kwargs)

        cur = self.db.cursor()
        cur.execute("SELECT key from clang_tidy_clean where key = ?", (cache_key, ))
        res = cur.fetchall()

        if len(res) > 0:
            if self.CACHE_DEBUG:
                print("Cache hit for", cache_key)
            return True
        else:
            if self.CACHE_DEBUG:
                print("Cache miss for", cache_key)
            return False

    def record(self, **kwargs):
        cache_key = hash_args(**kwargs)

        cur = self.db.cursor()
        if self.CACHE_DEBUG:
            print("Cache save for", cache_key)
        cur.execute("INSERT OR IGNORE INTO clang_tidy_clean values(?)", (cache_key, ))
        self.db.commit()

    def count(self):
        cur = self.db.cursor()
        cur.execute("SELECT count(*) FROM clang_tidy_clean")
        entries = cur.fetchall()
        return int(entries[0][0])

    def cleanup(self, limit):
        cur = self.db.cursor()

        cur.execute("SELECT key FROM clang_tidy_clean")
        entries = cur.fetchall()

        if len(entries) <= limit:
            return

        to_prune = len(entries) - limit

        # Randomly evict entries if cache has grown too large
        random.shuffle(entries)

        for cache_key in entries[:to_prune]:
            if self.CACHE_DEBUG:
                print("Pruning", cache_key[0])
            cur.execute("DELETE FROM clang_tidy_clean WHERE key = ?", (cache_key[0], ))
        self.db.commit()

def main(args = None): # pylint: disable=too-many-return-statements
    if args is None:
        args = sys.argv

    parser = optparse.OptionParser()

    parser.add_option('-j', '--jobs', action='store', type='int', default=multiprocessing.cpu_count() + 1,
                      help='set number of jobs to run (default %default)')
    parser.add_option('--quiet', action='store_true', default=False)
    parser.add_option('--fixit', action='store_true', default=False)
    parser.add_option('--build-dir', default='build')
    parser.add_option('--list-checks', action='store_true', default=False)
    parser.add_option('--regenerate-inline-config-file', action='store_true', default=False)
    parser.add_option('--only-changed-files', action='store_true', default=False)
    parser.add_option('--only-matching', metavar='REGEX', default='.*')
    parser.add_option('--take-file-list-from-stdin', action='store_true', default=False)
    parser.add_option('--export-fixes-dir', default=None)
    parser.add_option('--cache-db', metavar='DB',
                      help='Path to sqlite3 database file for caching',
                      default=os.getenv("BOTAN_CLANG_TIDY_CACHE"))

    # 100K entries with our current schema is about 15 Mb
    max_db_entries = int(os.getenv("BOTAN_CLANG_TIDY_CACHE_MAX_SIZE") or 100_000)

    (options, args) = parser.parse_args(args)

    if len(args) > 1:
        print("ERROR: Unknown extra arguments to run_clang_tidy.py")
        return 1

    if options.only_changed_files and options.take_file_list_from_stdin:
        print("Cannot use both --only-changed-files and --take-file-list-from-stdin", file=sys.stderr)
        return 1

    files_to_check = []
    if options.only_changed_files:
        changes = run_command(['git', 'diff', '--name-only', '-r', 'master'])

        files_to_check = []
        for file in changes.split():
            if file.endswith('.cpp') or file.endswith('.h'):
                files_to_check.append(os.path.basename(file))

        if len(files_to_check) == 0:
            print("No C++ files were modified vs master, skipping clang-tidy checks")
            return 0
    elif options.take_file_list_from_stdin:
        scan_all = False

        for line in sys.stdin:
            file = os.path.basename(line.strip())
            if file.endswith('.cpp') or file.endswith('.h'):
                files_to_check.append(file)
            elif file in ['run_clang_tidy.py']:
                scan_all = True

        if scan_all:
            # clear this to revert to disable filtering causing all files to be scanned
            files_to_check = []
        elif len(files_to_check) == 0:
            print("No C++ files were provided on stdin, skipping clang-tidy checks")
            return 0

    (compile_commands_file, compile_commands) = load_compile_commands(options.build_dir)

    check_config = create_check_option(enabled_checks, disabled_checks)

    if options.list_checks:
        print(run_command(['clang-tidy', '-list-checks', '-checks', check_config]), end='')
        return 0

    if options.regenerate_inline_config_file:
        render_clang_tidy_file('src', enabled_checks, disabled_checks)
        return 0

    pool = ThreadPool(options.jobs)

    cache = CacheDatabase(options.cache_db) if options.cache_db else None

    start_time = time.time()
    files_checked = 0

    file_matcher = re.compile(options.only_matching)

    clang_tidy_version = None

    if cache is not None:
        if not options.quiet:
            print("Using cache at %s with %d entries" % (options.cache_db, cache.count()))

        results = []

        clang_tidy_version = subprocess.run(['clang-tidy', '--version'],
                                            check=True,
                                            stdout=subprocess.PIPE).stdout.decode('utf-8')

        for info in compile_commands:
            results.append(pool.apply_async(preproc_file, (info, )))

        assert(len(results) == len(compile_commands))
        for (i, result) in enumerate(results):
            compile_commands[i]['preproc'] = result.get()

        if not options.quiet:
            print("Preprocessing/hashing %d files took %.02f sec" % (len(results), time.time() - start_time))
            sys.stdout.flush()

        start_time = time.time()

    results = []
    for info in compile_commands:
        file = info['file']

        if len(files_to_check) > 0 and os.path.basename(file) not in files_to_check:
            continue

        if file_matcher.search(file) is None:
            continue

        cache_hit = False
        if cache is not None:
            if cache.hit(config=check_config, source_file=info['preproc'], clang_tidy=clang_tidy_version):
                cache_hit = True

        if cache_hit and not options.quiet:
            print("Checked", info["file"], " (cached)")
            sys.stdout.flush()

        if not cache_hit:
            files_checked += 1

            results.append((info, pool.apply_async(
                run_clang_tidy,
                (compile_commands_file,
                 check_config,
                 file,
                 options))))

    fail_cnt = 0
    for (info, result) in results:
        success = result.get()

        if not options.quiet:
            print("Checked", info['file'])
            sys.stdout.flush()

        if success and cache is not None:
            cache.record(config=check_config, source_file=info['preproc'], clang_tidy=clang_tidy_version)

        if not success:
            fail_cnt += 1

    time_consumed = time.time() - start_time

    print("Checked %d files in %d seconds" % (files_checked, time_consumed))

    if cache is not None and max_db_entries > 0:
        cache.cleanup(max_db_entries)

    if fail_cnt == 0:
        return 0
    else:
        print("Found clang-tidy errors in %d files" % (fail_cnt))
        return 1

if __name__ == '__main__':
    sys.exit(main())



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import unittest

import botan3 as botan

from .wycheproof import WycheproofTests

_AEAD_CIPHER_ALIASES = {
    "AES-GCM": "AES",
    "AES-CCM": "AES",
    "AEAD-AES-SIV-CMAC": "AES-SIV-CMAC",
    "ARIA-GCM": "ARIA",
    "ARIA-CCM": "ARIA",
    "CAMELLIA-CCM": "Camellia",
    "SEED-GCM": "SEED",
    "SEED-CCM": "SEED",
    "SM4-GCM": "SM4",
    "SM4-CCM": "SM4",
}


def _aead_algorithm(
    algorithm: str, key_size_bits: int, tag_size_bits: int | None, nonce_len: int
) -> str:
    cipher = _AEAD_CIPHER_ALIASES.get(algorithm)
    if cipher is None:
        raise ValueError(f"Unsupported AEAD algorithm: {algorithm}")

    tag_len_bytes = tag_size_bits // 8 if tag_size_bits is not None else None

    if algorithm == "AEAD-AES-SIV-CMAC":
        aes_size = key_size_bits // 2
        return f"AES-{aes_size}/SIV"

    if cipher in ("AES", "ARIA", "Camellia"):
        cipher = f"{cipher}-{key_size_bits}"

    if algorithm.endswith("GCM"):
        suffix = f"/GCM({tag_len_bytes})" if tag_len_bytes is not None else "/GCM"
        return f"{cipher}{suffix}"

    if algorithm.endswith("CCM"):
        l_val = 15 - nonce_len
        if tag_len_bytes is None:
            return f"{cipher}/CCM({l_val})"
        return f"{cipher}/CCM({tag_len_bytes},{l_val})"

    raise ValueError(f"Unhandled AEAD algorithm: {algorithm}")


def _encrypt(aead: botan.SymmetricCipher, iv: bytes, aad: bytes, msg: bytes) -> bytes:
    aead.set_assoc_data(aad)
    aead.start(iv)
    return aead.finish(msg)


def _decrypt(
    aead: botan.SymmetricCipher, iv: bytes, aad: bytes, ct_tag: bytes
) -> bytes:
    aead.set_assoc_data(aad)
    aead.start(iv)
    return aead.finish(ct_tag)


class TestAEAD(WycheproofTests, unittest.TestCase):
    def input_files(self) -> list[str]:
        return [
            "aes_ccm_test.json",
            "aes_gcm_test.json",
            "aria_ccm_test.json",
            "aria_gcm_test.json",
            "camellia_ccm_test.json",
            "seed_ccm_test.json",
            "seed_gcm_test.json",
            "sm4_ccm_test.json",
            "sm4_gcm_test.json",
            "aead_aes_siv_cmac_test.json",
        ]

    def run_test(self, data: dict, group: dict, test: dict) -> None:
        algorithm = data["algorithm"]
        is_siv = algorithm == "AEAD-AES-SIV-CMAC"

        tag_size_bits = group.get("tagSize")
        key_size_bits = group.get("keySize")

        key = binascii.unhexlify(test["key"])
        iv = binascii.unhexlify(test["iv"])
        aad = binascii.unhexlify(test["aad"])
        msg = binascii.unhexlify(test["msg"])
        ct = binascii.unhexlify(test["ct"])
        tag = binascii.unhexlify(test["tag"])
        expected_ct_tag = (tag + ct) if is_siv else (ct + tag)

        algo = _aead_algorithm(algorithm, key_size_bits, tag_size_bits, len(iv))

        if test["result"] == "valid":
            enc = botan.SymmetricCipher(algo, True)
            enc.set_key(key)
            enc_out = _encrypt(enc, iv, aad, msg)
            self.assertEqual(enc_out, expected_ct_tag)

            dec = botan.SymmetricCipher(algo, False)
            dec.set_key(key)
            dec_out = _decrypt(dec, iv, aad, expected_ct_tag)
            self.assertEqual(dec_out, msg)
        elif test["result"] in ("invalid", "acceptable"):
            try:
                dec = botan.SymmetricCipher(algo, False)
                dec.set_key(key)
                dec_out = _decrypt(dec, iv, aad, expected_ct_tag)
                if test["result"] == "invalid":
                    self.assertNotEqual(dec_out, msg)
                if test["result"] == "acceptable":
                    self.assertEqual(dec_out, msg)
            except botan.BotanException:
                # may fail, because test is "invalid" or "acceptable"
                pass
        else:
            self.fail(f"Unknown test result: {test['result']}")



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import unittest

import botan3 as botan

from .wycheproof import WycheproofTests


def _from_hex(value: str) -> bytes:
    return binascii.unhexlify(value)


def _map_curve_name(wycheproof_curve: str) -> str:
    """Map Wycheproof curve name to Botan curve name."""
    mapping = {
        "brainpoolP224r1": "brainpool224r1",
        "brainpoolP256r1": "brainpool256r1",
        "brainpoolP384r1": "brainpool384r1",
        "brainpoolP512r1": "brainpool512r1",
    }
    return mapping.get(wycheproof_curve, wycheproof_curve)


class TestECDH(WycheproofTests, unittest.TestCase):
    def input_files(self) -> list[str]:
        return [
            "ecdh_secp224r1_test.json",
            "ecdh_secp256r1_test.json",
            "ecdh_secp256k1_test.json",
            "ecdh_secp384r1_test.json",
            "ecdh_secp521r1_test.json",
            "ecdh_brainpoolP224r1_test.json",
            "ecdh_brainpoolP256r1_test.json",
            "ecdh_brainpoolP384r1_test.json",
            "ecdh_brainpoolP512r1_test.json",
        ]

    def run_test(self, data: dict, group: dict, test: dict) -> None:
        curve = _map_curve_name(group["curve"])
        if not botan.ECGroup.supports_named_group(curve):
            self.skipTest(f"Curve {curve} not supported in this build")

        try:
            priv_value = botan.MPI(test["private"], radix=16)
            priv_key = botan.PrivateKey.load_ecdh(curve, priv_value)
        except botan.BotanException:
            if test["result"] != "valid":
                return
            raise

        try:
            pub_key = botan.PublicKey.load(_from_hex(test["public"]))
        except botan.BotanException:
            if test["result"] != "valid":
                return
            raise

        explicit_encoding = pub_key.used_explicit_encoding()
        if explicit_encoding:
            # Wycheproof encodes ECDH public keys as DER encoded EC points that
            # inherently allow an explicit encoding of the EC group. The Python
            # API can load such keys generically (PublicKey.load()), but the
            # PKKeyAgreement class expects the public value as SEC1 encoding
            # which does not allow the explicit encoding of the EC group.
            #
            # Therefore, PKKeyAgreement.agree() cannot detect the discrepancy in
            # the underlying EC groups and we need to handle this case manually.
            self.assertIn(test["result"], ("invalid", "acceptable"))
            return

        try:
            ka = botan.PKKeyAgreement(priv_key, "Raw")
            shared_secret = ka.agree(pub_key.to_raw(), 0, b"")
        except botan.BotanException:
            if test["result"] != "valid":
                return
            raise

        self.assertIn(
            test["result"],
            ("valid", "acceptable"),
            "Invalid test case produced a shared secret",
        )
        self.assertEqual(
            shared_secret,
            _from_hex(test["shared"]),
            "Shared secret does not match expected value",
        )



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import json
import os
import unittest
from abc import ABC, abstractmethod
from collections import Counter
from urllib.request import urlopen
from typing import BinaryIO
from pathlib import Path

import botan3 as botan


class WycheproofTextTestResult(unittest.TextTestResult):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._skip_reason_counts: Counter[str] = Counter()

    def addSkip(self, test, reason):
        self._skip_reason_counts[str(reason)] += 1
        super().addSkip(test, reason)

    def stopTestRun(self):
        super().stopTestRun()

        if not self._skip_reason_counts:
            return

        self.stream.writeln("")
        self.stream.writeln("Skipped tests summary (by reason):")

        for reason, count in self._skip_reason_counts.most_common():
            self.stream.writeln(f"  - {count}x: {reason}")


# When these Wycheproof tests are run via `python -m unittest`, the CLI uses
# `unittest.TextTestRunner` and its `resultclass`. Since all tests import this
# module, we can centrally augment the default output without changing each test.
if getattr(unittest.TextTestRunner, "resultclass", None) is unittest.TextTestResult:
    unittest.TextTestRunner.resultclass = WycheproofTextTestResult


class FixedOutputRNG(botan.RandomNumberGenerator):
    def __init__(self, entropy_pool: bytes = b""):
        """
        A random number generator that outputs a pre-determined sequence of bytes.
        This is exclusively useful for testing.
        """
        super().__init__(
            "custom", get_callback=self._get, add_entropy_callback=self._add_entropy
        )
        self._entropy_pool = entropy_pool

    def _get(self, length: int) -> bytes:
        if length > len(self._entropy_pool):
            raise ValueError("Not enough entropy in pool")
        entropy = self._entropy_pool[:length]
        self._entropy_pool = self._entropy_pool[length:]
        return entropy

    def _add_entropy(self, data: bytes) -> None:
        self._entropy_pool += data


# An RNG implementation called NullRNG that raises an exception if it is used
class NullRNG(botan.RandomNumberGenerator):
    def __init__(self):
        super().__init__(
            "custom", get_callback=self._get, add_entropy_callback=self._add_entropy
        )

    def _get(self, length: int) -> bytes:
        raise botan.BotanException("Unexpected request to get entropy from RNG", rc=-23)

    def _add_entropy(self, data: bytes) -> None:
        raise botan.BotanException("Unexpected request to add entropy to RNG", rc=-23)


class WycheproofTests(ABC):
    """
    Base mixin class for Wycheproof tests.

    This class is intended to be used as a mixin with unittest.TestCase.
    It explicitly relies on methods provided by unittest.TestCase

        class MyTest(WycheproofTests, unittest.TestCase):
            ...
    """

    @abstractmethod
    def input_files(self) -> list[str]:
        """
        Return the list of input files to test. Those files are expected to
        be in the `testvectors_v1` directory of the Wycheproof repository.
        """

    @abstractmethod
    def run_test(self, data: dict, group: dict, test: dict) -> None:
        """
        Run a single test.
        Args:
            data: The data for the test. This is the entire JSON file.
            group: The group for the test. This is the group of test vectors.
            test: The test itself. This is the individual test vector.
        """

    @staticmethod
    def _get_cached_or_downloaded_file(filename: str) -> BinaryIO:
        """
        Returns an open file handle to the JSON file with given filename.
        If WYCHEPROOF_TESTDATA_CACHE_DIR is set, the file is cached in that directory.
        """

        base_url = os.environ.get("WYCHEPROOF_TESTDATA_URL")
        if base_url is None:
            raise RuntimeError("Environment variable WYCHEPROOF_TESTDATA_URL not set")
        url = f"{base_url}/{filename}"

        # Check if the cache can be used
        cache_dir = os.environ.get("WYCHEPROOF_TESTDATA_CACHE_DIR")
        if cache_dir is not None:
            cache_path = Path(cache_dir)
            cache_path.mkdir(parents=True, exist_ok=True)
            cache_entry = cache_path / filename
            if not cache_entry.exists():
                with urlopen(url, timeout=30) as response:
                    with open(cache_entry, "wb") as f:
                        f.write(response.read())
            return open(cache_entry, mode="rb")
        return urlopen(url, timeout=30)

    @staticmethod
    def _read_datafile(filename: str) -> dict | None:
        """
        Reads the datafile named 'filename' from cache (if available) or downloads and caches it.
        Returns the loaded JSON content.
        """
        with WycheproofTests._get_cached_or_downloaded_file(filename) as f:
            return json.load(f) if f is not None else None

    def _validate_unittest_mixin(self) -> None:
        if not isinstance(self, unittest.TestCase):
            raise TypeError(
                f"{self.__class__.__name__} must be used as a mixin with unittest.TestCase"
            )

    def _wycheproof_subtest(self, test: dict, filename: str):
        self._validate_unittest_mixin()
        params = {}
        tc_id = test.get("tcId")
        if tc_id is not None:
            params["tcId"] = tc_id
        comment = test.get("comment")
        if comment:
            params["comment"] = comment
        flags = test.get("flags")
        if flags:
            params["flags"] = ",".join(flags)
        params["filename"] = filename
        return unittest.TestCase.subTest(self, **params)

    def test_wycheproof(self) -> None:
        self._validate_unittest_mixin()
        for filename in self.input_files():
            data = self._read_datafile(filename)
            for group in data["testGroups"]:
                for test in group["tests"]:
                    with self._wycheproof_subtest(test, filename):
                        self.run_test(data, group, test)



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import unittest

import botan3 as botan

from .wycheproof import WycheproofTests, FixedOutputRNG


def _from_hex(value: str) -> bytes:
    return binascii.unhexlify(value)


class TestMLKEM(WycheproofTests, unittest.TestCase):
    def input_files(self) -> list[str]:
        return [
            "mlkem_512_test.json",
            "mlkem_768_test.json",
            "mlkem_1024_test.json",
            "mlkem_512_keygen_seed_test.json",
            "mlkem_768_keygen_seed_test.json",
            "mlkem_1024_keygen_seed_test.json",
            "mlkem_512_encaps_test.json",
            "mlkem_768_encaps_test.json",
            "mlkem_1024_encaps_test.json",
        ]

    def run_test(self, _data: dict, group: dict, test: dict) -> None:
        mlkem_mode = group["parameterSet"]

        priv = None
        pub = None

        # load the private key, if a seed is provided
        if "seed" in test:
            seed = _from_hex(test["seed"])
            try:
                priv = botan.PrivateKey.load_ml_kem(mlkem_mode, seed)
            except botan.BotanException:
                if test["result"] == "invalid":
                    return
                raise
            pub = priv.get_public_key()

        # load and/or validate the public key
        if "ek" in test:
            expected_ek = _from_hex(test["ek"])

            # load the public key, if no private key was loaded
            if pub is None:
                try:
                    pub = botan.PublicKey.load_ml_kem(mlkem_mode, expected_ek)
                except botan.BotanException:
                    if test["result"] == "invalid":
                        return
                    raise

            self.assertEqual(pub.to_raw(), expected_ek)

        if not pub:
            raise ValueError("No public key available in this test vector")

        # reload and validate the private key
        if "dk" in test:
            try:
                expected_dk = _from_hex(test["dk"])
                priv2 = botan.PrivateKey.load_ml_kem(mlkem_mode, expected_dk)
                # TODO: currently we cannot export the expanded private key
                #       via the python API, we would need to be able to access
                #       ML_KEM_PrivateKey::private_key_bits_with_format()
                #
                # Hence, we only validate that the public key is the same.
                # and not that priv.to_expanded_raw() == expected_dk
                self.assertEqual(priv2.to_raw(), expected_dk)
                self.assertEqual(priv2.get_public_key().to_raw(), pub.to_raw())
            except botan.BotanException:
                if test["result"] in ("invalid", "acceptable"):
                    return
                raise

        if "c" in test and "K" in test:
            expected_k = _from_hex(test["K"])
            expected_c = _from_hex(test["c"])

            # encapsulation
            if "m" in test:
                rng = FixedOutputRNG(_from_hex(test["m"]))
                kem_e = botan.KemEncrypt(pub, "Raw")
                actual_k, actual_c = kem_e.create_shared_key(rng, b"", len(expected_k))
                self.assertEqual(actual_k, expected_k)
                self.assertEqual(actual_c, expected_c)

            # decapsulation
            if priv is not None:
                kem_d = botan.KemDecrypt(priv, "Raw")

                try:
                    actual_k = kem_d.decrypt_shared_key(
                        b"", len(expected_k), expected_c
                    )
                except botan.BotanException:
                    if test["result"] in ("invalid", "acceptable"):
                        return
                    raise

                if test["result"] == "valid":
                    self.assertEqual(actual_k, expected_k)
                elif test["result"] == "invalid":
                    self.assertNotEqual(actual_k, expected_k)
                elif test["result"] == "acceptable":
                    pass
                else:
                    self.fail(f"Unknown test result: {test['result']}")



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import unittest

import botan3 as botan

from .wycheproof import WycheproofTests

_MAC_ALGORITHMS = {
    "HMACSHA1": "HMAC(SHA-1)",
    "HMACSHA224": "HMAC(SHA-224)",
    "HMACSHA256": "HMAC(SHA-256)",
    "HMACSHA384": "HMAC(SHA-384)",
    "HMACSHA512": "HMAC(SHA-512)",
    "HMACSHA3-224": "HMAC(SHA-3(224))",
    "HMACSHA3-256": "HMAC(SHA-3(256))",
    "HMACSHA3-384": "HMAC(SHA-3(384))",
    "HMACSHA3-512": "HMAC(SHA-3(512))",
    "HMACSM3": "HMAC(SM3)",
    "SipHash-1-3": "SipHash(1,3)",
    "SipHash-2-4": "SipHash(2,4)",
    "SipHash-4-8": "SipHash(4,8)",
    "KMAC128": "KMAC-128",
    "KMAC256": "KMAC-256",
}


def _mac_algorithm(
    algorithm: str, key_size_bits: int | None, tag_size_bits: int | None
) -> str:
    if algorithm == "AES-GMAC":
        if key_size_bits is None:
            raise ValueError("AES-GMAC requires a key size")
        return f"GMAC(AES-{key_size_bits})"
    if algorithm.startswith("KMAC"):
        if tag_size_bits is None:
            raise ValueError("KMAC requires a tag size")
        return f"{_MAC_ALGORITHMS[algorithm]}({tag_size_bits})"
    return _MAC_ALGORITHMS[algorithm]


class TestMAC(WycheproofTests, unittest.TestCase):
    def input_files(self) -> list[str]:
        return [
            "hmac_sha1_test.json",
            "hmac_sha224_test.json",
            "hmac_sha256_test.json",
            "hmac_sha384_test.json",
            "hmac_sha3_224_test.json",
            "hmac_sha3_256_test.json",
            "hmac_sha3_384_test.json",
            "hmac_sha3_512_test.json",
            "hmac_sha512_test.json",
            "hmac_sm3_test.json",
            "siphash_1_3_test.json",
            "siphash_2_4_test.json",
            "siphash_4_8_test.json",
            "aes_gmac_test.json",
            "kmac128_no_customization_test.json",
            "kmac256_no_customization_test.json",
        ]

    def run_test(self, data: dict, group: dict, test: dict) -> None:
        key_size_bits = group.get("keySize")
        iv_size_bits = group.get("ivSize")
        tag_size_bits = group.get("tagSize")

        algo = _mac_algorithm(data["algorithm"], key_size_bits, tag_size_bits)

        mac = botan.MsgAuthCode(algo)
        mac.set_key(binascii.unhexlify(test["key"]))
        if iv_size_bits is not None:
            mac.set_nonce(binascii.unhexlify(test["iv"]))
        mac.update(binascii.unhexlify(test["msg"]))

        final_mac = mac.final()
        actual_mac = (
            final_mac[: tag_size_bits // 8] if tag_size_bits is not None else final_mac
        )

        expected = binascii.unhexlify(test["tag"])

        if test["result"] == "valid":
            self.assertEqual(actual_mac, expected)
        elif test["result"] == "invalid":
            self.assertNotEqual(actual_mac, expected)
        elif test["result"] == "acceptable":
            pass
        else:
            self.fail(f"Unknown test result: {test['result']}")



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import unittest

import botan3 as botan

from .wycheproof import WycheproofTests, NullRNG


def _from_hex(value: str) -> bytes:
    return binascii.unhexlify(value)


def _map_algorithm_to_mode(algorithm: str) -> str:
    """Map Wycheproof algorithm name to Botan ML-DSA mode name."""
    mapping = {
        "ML-DSA-44": "ML-DSA-4x4",
        "ML-DSA-65": "ML-DSA-6x5",
        "ML-DSA-87": "ML-DSA-8x7",
    }
    return mapping.get(algorithm, algorithm)


class TestMLDSA(WycheproofTests, unittest.TestCase):
    def input_files(self) -> list[str]:
        # TODO: enable *_noseed_*.json once support for expanded private keys is implemented
        return [
            "mldsa_44_sign_seed_test.json",
            # "mldsa_44_sign_noseed_test.json",
            "mldsa_44_verify_test.json",
            "mldsa_65_sign_seed_test.json",
            # "mldsa_65_sign_noseed_test.json",
            "mldsa_65_verify_test.json",
            "mldsa_87_sign_seed_test.json",
            # "mldsa_87_sign_noseed_test.json",
            "mldsa_87_verify_test.json",
        ]

    def run_test(self, data: dict, group: dict, test: dict) -> None:
        ctx = test.get("ctx")
        if ctx is not None and ctx != "":
            # Currently Botan doesn't support context (ctx), skip...
            self.skipTest("ML-DSA ctx not supported")

        if "msg" not in test or ("flags" in test and "Internal" in test["flags"]):
            # Currently Botan does not provide the "Sign_internal" interface of ML-DSA,
            # hence, signing without "msg" and only using "mu" isn't possible.
            self.skipTest("ML-DSA's Sign_internal interface is not exposed")

        mldsa_mode = _map_algorithm_to_mode(data.get("algorithm", ""))
        group_type = group.get("type")

        if group_type == "MlDsaSign":
            self._run_sign_test(mldsa_mode, group, test)
        elif group_type == "MlDsaVerify":
            self._run_verify_test(mldsa_mode, group, test)
        else:
            self.fail(f"Unknown test group type: {group_type}")

    def _run_sign_test(self, mldsa_mode: str, group: dict, test: dict) -> None:
        """Run a signing test (either seed or noseed variant)."""

        # Load the private key
        priv = None
        try:
            if "privateSeed" in group:
                priv = botan.PrivateKey.load_ml_dsa(
                    mldsa_mode, _from_hex(group["privateSeed"])
                )
            # TODO: implement loading from expanded private key and enable the relevant
            #       input files with *_noseed_*.json
        except botan.BotanException:
            if test["result"] == "invalid":
                return
            raise
        if priv is None:
            self.fail("No private key available for signing test")

        # Derive the public key and validate it against the expected public key
        pub = priv.get_public_key()
        if "publicKey" in group and group["publicKey"] is not None:
            self.assertEqual(
                pub.to_raw(),
                _from_hex(group["publicKey"]),
                "Deserialized public key does not match expected public key from test group",
            )

        # Perform signing
        try:
            signer = botan.PKSign(priv, "Deterministic")
            signer.update(_from_hex(test["msg"]))
            actual_sig = signer.finish(NullRNG())
        except botan.BotanException:
            if test["result"] == "invalid":
                return
            raise

        # Validate the generated signature
        self.assertEqual(
            actual_sig,
            _from_hex(test["sig"]),
            "Generated signature does not match expected signature in test vector",
        )

    def _run_verify_test(self, mldsa_mode: str, group: dict, test: dict) -> None:
        """Run a verification test."""
        # Load public key
        try:
            pub = botan.PublicKey.load_ml_dsa(mldsa_mode, _from_hex(group["publicKey"]))
        except botan.BotanException:
            if test["result"] == "invalid":
                return
            raise

        # Perform verification
        verifier = botan.PKVerify(pub, "")
        verifier.update(_from_hex(test["msg"]))
        valid = verifier.check_signature(_from_hex(test["sig"]))

        # Validate the verification result
        self.assertEqual(
            valid,
            test["result"] == "valid",
            "Signature should be valid"
            if test["result"] == "valid"
            else "Signature should be invalid",
        )



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import unittest

import botan3 as botan

from .wycheproof import WycheproofTests


def _from_hex(value: str) -> bytes:
    return binascii.unhexlify(value)


def _map_curve_name(wycheproof_curve: str) -> str:
    """Map Wycheproof curve name to Botan curve name."""
    mapping = {
        "brainpoolP224r1": "brainpool224r1",
        "brainpoolP256r1": "brainpool256r1",
        "brainpoolP320r1": "brainpool320r1",
        "brainpoolP384r1": "brainpool384r1",
        "brainpoolP512r1": "brainpool512r1",
    }
    return mapping.get(wycheproof_curve, wycheproof_curve)


def _map_hash_name(wycheproof_hash: str) -> str | None:
    """Map Wycheproof hash name to Botan hash name.
    Returns None for unsupported hashes.
    """
    mapping = {
        "SHA-224": "SHA-224",
        "SHA-256": "SHA-256",
        "SHA-384": "SHA-384",
        "SHA-512": "SHA-512",
        "SHA3-224": "SHA-3(224)",
        "SHA3-256": "SHA-3(256)",
        "SHA3-384": "SHA-3(384)",
        "SHA3-512": "SHA-3(512)",
        "SHAKE128": "SHAKE-128(256)",
        "SHAKE256": "SHAKE-256(512)",
    }
    return mapping.get(wycheproof_hash)


class TestECDSA(WycheproofTests, unittest.TestCase):
    def input_files(self) -> list[str]:
        return [
            "ecdsa_brainpoolP224r1_sha224_p1363_test.json",
            "ecdsa_brainpoolP224r1_sha224_test.json",
            "ecdsa_brainpoolP224r1_sha3_224_test.json",
            "ecdsa_brainpoolP256r1_sha256_p1363_test.json",
            "ecdsa_brainpoolP256r1_sha256_test.json",
            "ecdsa_brainpoolP256r1_sha3_256_test.json",
            "ecdsa_brainpoolP320r1_sha384_p1363_test.json",
            "ecdsa_brainpoolP320r1_sha384_test.json",
            "ecdsa_brainpoolP320r1_sha3_384_test.json",
            "ecdsa_brainpoolP384r1_sha384_p1363_test.json",
            "ecdsa_brainpoolP384r1_sha384_test.json",
            "ecdsa_brainpoolP384r1_sha3_384_test.json",
            "ecdsa_brainpoolP512r1_sha3_512_test.json",
            "ecdsa_brainpoolP512r1_sha512_p1363_test.json",
            "ecdsa_brainpoolP512r1_sha512_test.json",
            "ecdsa_secp160k1_sha256_p1363_test.json",
            "ecdsa_secp160k1_sha256_test.json",
            "ecdsa_secp160r1_sha256_p1363_test.json",
            "ecdsa_secp160r1_sha256_test.json",
            "ecdsa_secp160r2_sha256_p1363_test.json",
            "ecdsa_secp160r2_sha256_test.json",
            "ecdsa_secp192k1_sha256_p1363_test.json",
            "ecdsa_secp192k1_sha256_test.json",
            "ecdsa_secp192r1_sha256_p1363_test.json",
            "ecdsa_secp192r1_sha256_test.json",
            "ecdsa_secp224k1_sha224_p1363_test.json",
            "ecdsa_secp224k1_sha224_test.json",
            "ecdsa_secp224k1_sha256_p1363_test.json",
            "ecdsa_secp224k1_sha256_test.json",
            "ecdsa_secp224r1_sha224_p1363_test.json",
            "ecdsa_secp224r1_sha224_test.json",
            "ecdsa_secp224r1_sha256_p1363_test.json",
            "ecdsa_secp224r1_sha256_test.json",
            "ecdsa_secp224r1_sha3_224_test.json",
            "ecdsa_secp224r1_sha3_256_test.json",
            "ecdsa_secp224r1_sha3_512_test.json",
            "ecdsa_secp224r1_sha512_p1363_test.json",
            "ecdsa_secp224r1_sha512_test.json",
            "ecdsa_secp224r1_shake128_p1363_test.json",
            "ecdsa_secp224r1_shake128_test.json",
            "ecdsa_secp256k1_sha256_bitcoin_test.json",
            "ecdsa_secp256k1_sha256_p1363_test.json",
            "ecdsa_secp256k1_sha256_test.json",
            "ecdsa_secp256k1_sha3_256_test.json",
            "ecdsa_secp256k1_sha3_512_test.json",
            "ecdsa_secp256k1_sha512_p1363_test.json",
            "ecdsa_secp256k1_sha512_test.json",
            "ecdsa_secp256k1_shake128_p1363_test.json",
            "ecdsa_secp256k1_shake128_test.json",
            "ecdsa_secp256k1_shake256_p1363_test.json",
            "ecdsa_secp256k1_shake256_test.json",
            "ecdsa_secp256r1_sha256_p1363_test.json",
            "ecdsa_secp256r1_sha256_test.json",
            "ecdsa_secp256r1_sha3_256_test.json",
            "ecdsa_secp256r1_sha3_512_test.json",
            "ecdsa_secp256r1_sha512_p1363_test.json",
            "ecdsa_secp256r1_sha512_test.json",
            "ecdsa_secp256r1_shake128_p1363_test.json",
            "ecdsa_secp256r1_shake128_test.json",
            "ecdsa_secp384r1_sha256_test.json",
            "ecdsa_secp384r1_sha384_p1363_test.json",
            "ecdsa_secp384r1_sha384_test.json",
            "ecdsa_secp384r1_sha3_384_test.json",
            "ecdsa_secp384r1_sha3_512_test.json",
            "ecdsa_secp384r1_sha512_p1363_test.json",
            "ecdsa_secp384r1_sha512_test.json",
            "ecdsa_secp384r1_shake256_p1363_test.json",
            "ecdsa_secp384r1_shake256_test.json",
            "ecdsa_secp521r1_sha3_512_test.json",
            "ecdsa_secp521r1_sha512_p1363_test.json",
            "ecdsa_secp521r1_sha512_test.json",
            "ecdsa_secp521r1_shake256_p1363_test.json",
            "ecdsa_secp521r1_shake256_test.json",
        ]

    def run_test(self, data: dict, group: dict, test: dict) -> None:
        botan_hash = _map_hash_name(group["sha"])
        curve = _map_curve_name(group["publicKey"]["curve"])
        if not botan.ECGroup.supports_named_group(curve):
            self.skipTest(f"Curve {curve} not supported in this build")

        # Load public key from DER
        try:
            pub_key = botan.PublicKey.load(_from_hex(group["publicKeyDer"]))
        except botan.BotanException:
            if test["result"] != "valid":
                return
            raise

        # Determine signature format from group type
        group_type = group["type"]
        if group_type == "EcdsaBitcoinVerify":
            # Botan currently does not prevent signature malleability via
            # low-s normalization.
            self.skipTest("Bitcoin variant of ECDSA is not supported")

        if group_type in ("EcdsaVerify", "EcdsaBitcoinVerify"):
            use_der = True
        elif group_type == "EcdsaP1363Verify":
            use_der = False
        else:
            self.fail(f"Unknown test group type: {group_type}")

        # Perform verification
        try:
            verifier = botan.PKVerify(pub_key, botan_hash, der=use_der)
            verifier.update(_from_hex(test["msg"]))
            valid = verifier.check_signature(_from_hex(test["sig"]))
        except botan.BotanException:
            if test["result"] != "valid":
                return
            raise

        # Validate the verification result
        self.assertEqual(
            valid,
            test["result"] == "valid",
            "Signature should be valid"
            if test["result"] == "valid"
            else "Signature should be invalid",
        )



"""
(C) 2026 Jack Lloyd
(C) 2026 René Meusel, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)
"""

import binascii
import unittest
import botan3 as botan

from .wycheproof import WycheproofTests


_HKDF_ALIASES = {
    "HKDF-SHA-1": "HKDF(SHA-1)",
    "HKDF-SHA-256": "HKDF(SHA-256)",
    "HKDF-SHA-384": "HKDF(SHA-384)",
    "HKDF-SHA-512": "HKDF(SHA-512)",
}


class TestHKDF(WycheproofTests, unittest.TestCase):
    def input_files(self) -> list[str]:
        return [
            "hkdf_sha1_test.json",
            "hkdf_sha256_test.json",
            "hkdf_sha384_test.json",
            "hkdf_sha512_test.json",
        ]

    def run_test(self, data: dict, _group: dict, test: dict) -> None:
        algo = _HKDF_ALIASES.get(data["algorithm"])
        if algo is None:
            raise ValueError(f"Unsupported HKDF algorithm: {data['algorithm']}")

        ikm = binascii.unhexlify(test["ikm"])
        salt = binascii.unhexlify(test["salt"])
        info = binascii.unhexlify(test["info"])
        size = test["size"]
        expected = binascii.unhexlify(test["okm"])

        try:
            actual = botan.kdf(algo, ikm, size, salt, info)
        except botan.BotanException:
            if test["result"] in ("invalid", "acceptable"):
                return
            raise

        if test["result"] == "valid":
            self.assertEqual(actual, expected)
        elif test["result"] == "invalid":
            self.assertNotEqual(actual, expected)
        elif test["result"] == "acceptable":
            pass
        else:
            self.fail(f"Unknown test result: {test['result']}")
